"""
Auto-generated class for DHCP
"""
from .GWHost import GWHost

from . import client_support


class DHCP(object):
    """
    auto-generated. don't touch.
    """

    @staticmethod
    def create(hosts, nameservers=None):
        """
        :type hosts: list[GWHost]
        :type nameservers: list[str]
        :rtype: DHCP
        """

        return DHCP(
            hosts=hosts,
            nameservers=nameservers,
        )

    def __init__(self, json=None, **kwargs):
        if json is None and not kwargs:
            raise ValueError('No data or kwargs present')

        class_name = 'DHCP'
        create_error = '{cls}: unable to create {prop} from value: {val}: {err}'
        required_error = '{cls}: missing required property {prop}'

        data = json or kwargs

        property_name = 'hosts'
        val = data.get(property_name)
        if val is not None:
            datatypes = [GWHost]
            try:
                self.hosts = client_support.list_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))
        else:
            raise ValueError(required_error.format(cls=class_name, prop=property_name))

        property_name = 'nameservers'
        val = data.get(property_name)
        if val is not None:
            datatypes = [str]
            try:
                self.nameservers = client_support.list_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))

    def __str__(self):
        return self.as_json(indent=4)

    def as_json(self, indent=0):
        return client_support.to_json(self, indent=indent)

    def as_dict(self):
        return client_support.to_dict(self)



from ..healthcheck import IPMIHealthCheck

descr = """
Checks the power redundancy of a node using IPMItool.
Result will be shown in the "Hardware" section of the Grid Portal / Status Overview / Node Status page.
"""


class PowerSupply(IPMIHealthCheck):
    def __init__(self, node):
        resource = '/nodes/{}'.format(node.name)
        super().__init__(id='pw-supply', name='Power Supply', category="Hardware", resource=resource)
        self.node = node
        self.ps_errmsgs = [
            "Power Supply AC lost",
            "Failure detected",
            "Predictive failure",
            "AC lost or out-of-range",
            "AC out-of-range, but present",
            "Config Erro",
            "Power Supply Inactive"]

    def run(self, container):
        ps_errmsgs = [x.lower() for x in self.ps_errmsgs if x.strip()]
        linehaserrmsg = lambda line: any([x in line.lower() for x in ps_errmsgs])

        out = self.execute_ipmi(container, """ipmitool -c sdr type "Power Supply" """)
        if out:

            # SAMPLE 1:
            # root@du-conv-3-01:~# ipmitool -c sdr type "Power Supply"
            # PS1 Status , C8h , ok , 10.1 , Presence detected
            # PS2 Status,C9h , ok , 10.2 , Presence detected

            # SAMPLE 2:
            # root@stor-04:~# ipmitool -c sdr type "Power Supply"
            # PSU1_Status , DEh , ok , 10.1 , Presence detected
            # PSU2_Status , DFh , ns , 10.2 , No Reading
            # PSU3_Status , E0h , ok , 10.3 , Presence detected
            # PSU4_Status , E1h , ns , 10.4 , No Reading
            # PSU Redundancy , E6h , ok , 21.1 , Fully Redundant

            # SAMPLE 3:
            # root@stor-01:~# ipmitool -c sdr type "Power Supply"
            # PSU1_Status , DEh , ok , 10.1 , Presence detected ,  Power Supply AC lost
            # PSU2_Status , DFh , ns , 10.2 , No Reading
            # PSU3_Status , E0h , ok , 10.3 , Presence detected
            # PSU4_Status , E1h , ok , 10.4 , Presence detected
            # PSU Redundancy , E6h , ok , 21.1 , Redundancy Lost
            # PSU Alert , 16h , ns , 208.1 , Event-Only

            psu_redun_in_out = "PSU Redundancy".lower() in out.lower()
            is_fully_redundant = True if "fully redundant" in out.lower() else False
            for line in out.splitlines():
                if "status" in line.lower():
                    parts = [part.strip() for part in line.split(",")]
                    id_, presence = parts[0], parts[-1]
                    id_ = id_.strip("Status").strip("_").strip()  # clean the power supply name.
                    if linehaserrmsg(line):
                        if psu_redun_in_out and is_fully_redundant:
                            self.add_message(id=id_, status='SKIPPED', text="Power redundancy problem on %s (%s)" % (id_, presence))
                        else:
                            self.add_message(id=id_, status='WARNING', text="Power redundancy problem on %s (%s)" % (id_, presence))
                    else:
                        self.add_message(id=id_, status='OK', text="Power supply %s is OK" % id_)
        else:
            self.add_message(id="SKIPPED", status='SKIPPED', text="No data for Power Supplies")



from js9 import j
from ..healthcheck import HealthCheckRun

descr = """
Checks average memory and CPU usage/load. If average per hour is higher than expected an error condition is thrown.

For both memory and CPU usage throws WARNING if more than 80% used and throws ERROR if more than 95% used.

Result will be shown in the "System Load" section of the Grid Portal / Status Overview / Node Status page.
"""


class Memory(HealthCheckRun):
    def __init__(self, node):
        resource = '/nodes/{}'.format(node.name)
        super().__init__('memory', 'Memory', 'System Load', resource)
        self.node = node

    def run(self):
        total_mem = self.node.client.info.mem()['total']/(1024*1024)
        mem_history = self.node.client.aggregator.query('machine.memory.ram.available').get('machine.memory.ram.available', {}).get('history', {})

        if '3600' not in mem_history:
            self.add_message('MEMORY', 'WARNING', 'Average memory load is not collected yet')
        else:
            avg_available_mem = mem_history['3600'][-1]['avg']
            avg_used_mem = total_mem - avg_available_mem
            avg_mem_percent = avg_used_mem/float(total_mem) * 100
            self.add_message(**get_message('memory', avg_mem_percent))


class CPU(HealthCheckRun):
    def __init__(self, node):
        resource = '/nodes/{}'.format(node.name)
        super().__init__('CPU', 'CPU', 'System Load', resource)
        self.node = node

    def run(self):
        cpu_percent = 0
        count = 0
        cpu_usage = self.node.client.aggregator.query('machine.CPU.percent')
        for cpu, data in cpu_usage.items():
            if '3600' not in data['history']:
                continue
            cpu_percent += (data['history']['3600'][-1]['avg'])
            count += 1

        if count == 0:
            self.add_message('CPU', 'WARNING', 'Average CPU load is not collected yet')
        else:
            cpu_avg = cpu_percent / float(count)
            self.add_message(**get_message('cpu', cpu_avg))


def get_message(type_, percent):
    message = {
        'id': type_.upper(),
        'status': 'OK',
        'text': r'Average %s load during last hour was: %.2f%%' % (type_.upper(), percent),
    }

    if percent > 95:
        message['status'] = 'ERROR'
        message['text'] = r'Average %s load during last hour was too high' % (type_.upper())
    elif percent > 80:
        message['status'] = 'WARNING'
        message['text'] = r'Average %s load during last hour was too high' % (type_.upper())

    return message



from enum import Enum


class EnumStoragePoolCreateDataProfile(Enum):
    raid0 = "raid0"
    raid1 = "raid1"
    raid5 = "raid5"
    raid6 = "raid6"
    raid10 = "raid10"
    dup = "dup"
    single = "single"



import re
import os
import datetime
from ..healthcheck import HealthCheckRun

descr = """
Rotate know log files if their size hit 10G or more
"""


class RotateLogs(HealthCheckRun):
    def __init__(self, node):
        resource = '/nodes/{}'.format(node.name)
        super().__init__('log-rotator', 'Log Rotator', 'System Load', resource)
        self.node = node

    def run(self, locations=['/var/log'], limit=10):
        message = {
            'id': 'log-rotator',
            'status': 'OK',
            'text': 'Logs did not reach {limit}G'.format(limit=limit)
        }

        if "/var/log" not in locations:
            locations.append("/var/log")
        logs = []
        try:
            # Get total size for all log files
            log_size = get_log_size(self.node, locations)

            # Rotate logs if they are larger than the limit
            if log_size/(1024 * 1024 * 1024) >= limit:  # convert bytes to GIGA
                # Rotate core.log

                for location in locations:
                    # Get Files for this location
                    location_files = get_files(self.node, location, [])
                    logs.extend(location_files)

                for file_path in logs:
                    if file_path == '/var/log/core.log':
                        continue
                    # Delete old rotated files to free up some space
                    # match file.*.date.time
                    if re.match(".*\d{8}-\d{6}", file_path):
                        self.node.client.filesystem.remove(file_path)
                    else:
                        new_path = "%s.%s" % (file_path, datetime.datetime.now().strftime('%Y%m%d-%H%M%S'))
                        # Rotate the new logs
                        self.node.client.filesystem.move(file_path, new_path)
                        # Create a new EMPTY file with the same name
                        fd = self.node.client.filesystem.open(file_path, 'x')
                        self.node.client.filesystem.close(fd)
                self.node.client.logger.reopen()
                message['text'] = 'Logs cleared'
        except Exception as e:
            message['text'] = "Error happened, Can not clear logs"
            message['status'] = "ERROR"

        self.add_message(**message)


def get_files(node, location, files=[]):
    for item in node.client.filesystem.list(location):
        if not item['is_dir']:
            files.append(os.path.join(location, item['name']))
        else:
            files = get_files(node, os.path.join(location, item['name']), files)
    return files


def get_log_size(node, locations):
    size = 0
    for location in locations:
        items = node.client.filesystem.list(location)
        for item in items:
            size += item['size']
    return size



import re
from ..healthcheck import HealthCheckRun

descr = """
Monitors if a network bond (if there is one) has both (or more) interfaces properly active.
"""


class NetworkBond(HealthCheckRun):
    def __init__(self, node):
        resource = '/nodes/{}'.format(node.name)
        super().__init__('network-bond', 'Network Bond Check', 'Hardware', resource)
        self.node = node

    def run(self):
        ovs = "{}_ovs".format(self.node.name)
        try:
            container = self.node.containers.get(ovs)
        except LookupError:
            # no ovs configured nothing to report on
            return
        jobresult = container.client.system('ovs-appctl bond/show').get()
        if jobresult.state == 'ERROR':
            return
        output = jobresult.stdout
        bonds = []
        bond = {}
        for match in re.finditer('(?:---- bond-(?P<bondname>\w+) ----)?.+?\n(?:slave (?:(?P<slavename>\w+): (?P<state>\w+)))', output, re.DOTALL):
            groups = match.groupdict()
            slave = {'name': groups['slavename'], 'state': groups['state']}
            if groups['bondname']:
                if bond:
                    bonds.append(bond)
                bond = {'name': groups['bondname']}
            bond.setdefault('slaves', []).append(slave)
        if bond:
            bonds.append(bond)

        for bond in bonds:
            badslaves = []
            for slave in bond['slaves']:
                if slave['state'] != 'enabled':
                    badslaves.append(slave['name'])
            state = 'OK'
            if badslaves:
                msg = 'Bond: {} has problems with slaves {}'.format(bond['name'], ', '.join(badslaves))
                state = 'ERROR' if len(badslaves) == len(bond['slaves']) else 'WARNING'
            else:
                msg = 'Bond: {}, all slave are ok'.format(bond['name'])

            self.add_message(bond, state, msg)



"""
Auto-generated class for MemInfo
"""

from . import client_support


class MemInfo(object):
    """
    auto-generated. don't touch.
    """

    @staticmethod
    def create(active, available, buffers, cached, free, inactive, total, used, usedPercent, wired):
        """
        :type active: int
        :type available: int
        :type buffers: int
        :type cached: int
        :type free: int
        :type inactive: int
        :type total: int
        :type used: int
        :type usedPercent: float
        :type wired: int
        :rtype: MemInfo
        """

        return MemInfo(
            active=active,
            available=available,
            buffers=buffers,
            cached=cached,
            free=free,
            inactive=inactive,
            total=total,
            used=used,
            usedPercent=usedPercent,
            wired=wired,
        )

    def __init__(self, json=None, **kwargs):
        if json is None and not kwargs:
            raise ValueError('No data or kwargs present')

        class_name = 'MemInfo'
        create_error = '{cls}: unable to create {prop} from value: {val}: {err}'
        required_error = '{cls}: missing required property {prop}'

        data = json or kwargs

        property_name = 'active'
        val = data.get(property_name)
        if val is not None:
            datatypes = [int]
            try:
                self.active = client_support.val_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))
        else:
            raise ValueError(required_error.format(cls=class_name, prop=property_name))

        property_name = 'available'
        val = data.get(property_name)
        if val is not None:
            datatypes = [int]
            try:
                self.available = client_support.val_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))
        else:
            raise ValueError(required_error.format(cls=class_name, prop=property_name))

        property_name = 'buffers'
        val = data.get(property_name)
        if val is not None:
            datatypes = [int]
            try:
                self.buffers = client_support.val_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))
        else:
            raise ValueError(required_error.format(cls=class_name, prop=property_name))

        property_name = 'cached'
        val = data.get(property_name)
        if val is not None:
            datatypes = [int]
            try:
                self.cached = client_support.val_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))
        else:
            raise ValueError(required_error.format(cls=class_name, prop=property_name))

        property_name = 'free'
        val = data.get(property_name)
        if val is not None:
            datatypes = [int]
            try:
                self.free = client_support.val_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))
        else:
            raise ValueError(required_error.format(cls=class_name, prop=property_name))

        property_name = 'inactive'
        val = data.get(property_name)
        if val is not None:
            datatypes = [int]
            try:
                self.inactive = client_support.val_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))
        else:
            raise ValueError(required_error.format(cls=class_name, prop=property_name))

        property_name = 'total'
        val = data.get(property_name)
        if val is not None:
            datatypes = [int]
            try:
                self.total = client_support.val_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))
        else:
            raise ValueError(required_error.format(cls=class_name, prop=property_name))

        property_name = 'used'
        val = data.get(property_name)
        if val is not None:
            datatypes = [int]
            try:
                self.used = client_support.val_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))
        else:
            raise ValueError(required_error.format(cls=class_name, prop=property_name))

        property_name = 'usedPercent'
        val = data.get(property_name)
        if val is not None:
            datatypes = [float]
            try:
                self.usedPercent = client_support.val_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))
        else:
            raise ValueError(required_error.format(cls=class_name, prop=property_name))

        property_name = 'wired'
        val = data.get(property_name)
        if val is not None:
            datatypes = [int]
            try:
                self.wired = client_support.val_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))
        else:
            raise ValueError(required_error.format(cls=class_name, prop=property_name))

    def __str__(self):
        return self.as_json(indent=4)

    def as_json(self, indent=0):
        return client_support.to_json(self, indent=indent)

    def as_dict(self):
        return client_support.to_dict(self)



"""
Auto-generated class for VdiskCreate
"""
from .EnumVdiskCreateType import EnumVdiskCreateType

from . import client_support


class VdiskCreate(object):
    """
    auto-generated. don't touch.
    """

    @staticmethod
    def create(blocksize, id, size, storagecluster, type, backupStoragecluster=None, readOnly=None, templatevdisk=None, tlogStoragecluster=None):
        """
        :type backupStoragecluster: str
        :type blocksize: int
        :type id: str
        :type readOnly: bool
        :type size: int
        :type storagecluster: str
        :type templatevdisk: str
        :type tlogStoragecluster: str
        :type type: EnumVdiskCreateType
        :rtype: VdiskCreate
        """

        return VdiskCreate(
            backupStoragecluster=backupStoragecluster,
            blocksize=blocksize,
            id=id,
            readOnly=readOnly,
            size=size,
            storagecluster=storagecluster,
            templatevdisk=templatevdisk,
            tlogStoragecluster=tlogStoragecluster,
            type=type,
        )

    def __init__(self, json=None, **kwargs):
        if json is None and not kwargs:
            raise ValueError('No data or kwargs present')

        class_name = 'VdiskCreate'
        create_error = '{cls}: unable to create {prop} from value: {val}: {err}'
        required_error = '{cls}: missing required property {prop}'

        data = json or kwargs

        property_name = 'backupStoragecluster'
        val = data.get(property_name)
        if val is not None:
            datatypes = [str]
            try:
                self.backupStoragecluster = client_support.val_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))

        property_name = 'blocksize'
        val = data.get(property_name)
        if val is not None:
            datatypes = [int]
            try:
                self.blocksize = client_support.val_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))
        else:
            raise ValueError(required_error.format(cls=class_name, prop=property_name))

        property_name = 'id'
        val = data.get(property_name)
        if val is not None:
            datatypes = [str]
            try:
                self.id = client_support.val_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))
        else:
            raise ValueError(required_error.format(cls=class_name, prop=property_name))

        property_name = 'readOnly'
        val = data.get(property_name)
        if val is not None:
            datatypes = [bool]
            try:
                self.readOnly = client_support.val_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))

        property_name = 'size'
        val = data.get(property_name)
        if val is not None:
            datatypes = [int]
            try:
                self.size = client_support.val_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))
        else:
            raise ValueError(required_error.format(cls=class_name, prop=property_name))

        property_name = 'storagecluster'
        val = data.get(property_name)
        if val is not None:
            datatypes = [str]
            try:
                self.storagecluster = client_support.val_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))
        else:
            raise ValueError(required_error.format(cls=class_name, prop=property_name))

        property_name = 'templatevdisk'
        val = data.get(property_name)
        if val is not None:
            datatypes = [str]
            try:
                self.templatevdisk = client_support.val_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))

        property_name = 'tlogStoragecluster'
        val = data.get(property_name)
        if val is not None:
            datatypes = [str]
            try:
                self.tlogStoragecluster = client_support.val_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))

        property_name = 'type'
        val = data.get(property_name)
        if val is not None:
            datatypes = [EnumVdiskCreateType]
            try:
                self.type = client_support.val_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))
        else:
            raise ValueError(required_error.format(cls=class_name, prop=property_name))

    def __str__(self):
        return self.as_json(indent=4)

    def as_json(self, indent=0):
        return client_support.to_json(self, indent=indent)

    def as_dict(self):
        return client_support.to_dict(self)



"""
Auto-generated class for Process
"""
from .CPUStats import CPUStats

from . import client_support


class Process(object):
    """
    auto-generated. don't touch.
    """

    @staticmethod
    def create(cmdline, cpu, pid, rss, swap, vms):
        """
        :type cmdline: str
        :type cpu: CPUStats
        :type pid: int
        :type rss: int
        :type swap: int
        :type vms: int
        :rtype: Process
        """

        return Process(
            cmdline=cmdline,
            cpu=cpu,
            pid=pid,
            rss=rss,
            swap=swap,
            vms=vms,
        )

    def __init__(self, json=None, **kwargs):
        if json is None and not kwargs:
            raise ValueError('No data or kwargs present')

        class_name = 'Process'
        create_error = '{cls}: unable to create {prop} from value: {val}: {err}'
        required_error = '{cls}: missing required property {prop}'

        data = json or kwargs

        property_name = 'cmdline'
        val = data.get(property_name)
        if val is not None:
            datatypes = [str]
            try:
                self.cmdline = client_support.val_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))
        else:
            raise ValueError(required_error.format(cls=class_name, prop=property_name))

        property_name = 'cpu'
        val = data.get(property_name)
        if val is not None:
            datatypes = [CPUStats]
            try:
                self.cpu = client_support.val_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))
        else:
            raise ValueError(required_error.format(cls=class_name, prop=property_name))

        property_name = 'pid'
        val = data.get(property_name)
        if val is not None:
            datatypes = [int]
            try:
                self.pid = client_support.val_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))
        else:
            raise ValueError(required_error.format(cls=class_name, prop=property_name))

        property_name = 'rss'
        val = data.get(property_name)
        if val is not None:
            datatypes = [int]
            try:
                self.rss = client_support.val_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))
        else:
            raise ValueError(required_error.format(cls=class_name, prop=property_name))

        property_name = 'swap'
        val = data.get(property_name)
        if val is not None:
            datatypes = [int]
            try:
                self.swap = client_support.val_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))
        else:
            raise ValueError(required_error.format(cls=class_name, prop=property_name))

        property_name = 'vms'
        val = data.get(property_name)
        if val is not None:
            datatypes = [int]
            try:
                self.vms = client_support.val_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))
        else:
            raise ValueError(required_error.format(cls=class_name, prop=property_name))

    def __str__(self):
        return self.as_json(indent=4)

    def as_json(self, indent=0):
        return client_support.to_json(self, indent=indent)

    def as_dict(self):
        return client_support.to_dict(self)



import json
from io import BytesIO

import logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class Containers:
    def __init__(self, node):
        self.node = node

    def list(self):
        containers = []
        for container in self.node.client.container.list().values():
            try:
                containers.append(Container.from_containerinfo(container, self.node))
            except ValueError:
                # skip containers withouth tags
                pass
        return containers

    def get(self, name):
        containers = list(self.node.client.container.find(name).values())
        if not containers:
            raise LookupError("Could not find container with name {}".format(name))
        if len(containers) > 1:
            raise LookupError("Found more than one containter with name {}".format(name))
        return Container.from_containerinfo(containers[0], self.node)

    def create(self, name, flist, hostname=None, mounts=None, nics=None,
               host_network=False, ports=None, storage=None, init_processes=None, privileged=False):
        logger.debug("create container %s", name)
        container = Container(name, self.node, flist, hostname, mounts, nics,
                              host_network, ports, storage, init_processes, privileged)
        container.start()
        return container


class Container:
    """G8SO Container"""

    def __init__(self, name, node, flist, hostname=None, mounts=None, nics=None,
                 host_network=False, ports=None, storage=None, init_processes=None,
                 privileged=False, identity=None):
        """
        TODO: write doc string
        filesystems: dict {filesystemObj: target}
        """
        self.name = name
        self.node = node
        self.mounts = mounts or {}
        self.hostname = hostname
        self.flist = flist
        self.ports = ports or {}
        self.nics = nics or []
        self.host_network = host_network
        self.storage = storage
        self.init_processes = init_processes or []
        self._client = None
        self.privileged = privileged
        self.identity = identity

        self._ays = None
        for nic in self.nics:
            nic.pop('token', None)
            if nic.get('config', {}).get('gateway', ''):
                nic['monitor'] = True

    @classmethod
    def from_containerinfo(cls, containerinfo, node):
        logger.debug("create container from info")
        arguments = containerinfo['container']['arguments']
        if not arguments['tags']:
            # we don't deal with tagless containers
            raise ValueError("Could not load containerinfo withouth tags")
        return cls(arguments['tags'][0],
                   node,
                   arguments['root'],
                   arguments['hostname'],
                   arguments['mount'],
                   arguments['nics'],
                   arguments['host_network'],
                   arguments['port'],
                   arguments['storage'],
                   arguments['privileged'],
                   arguments['identity'])

    @classmethod
    def from_ays(cls, service, password=None):
        logger.debug("create container from service (%s)", service)
        from .Node import Node
        node = Node.from_ays(service.parent, password)
        ports = {}
        for portmap in service.model.data.ports:
            source, dest = portmap.split(':')
            ports[int(source)] = int(dest)
        nics = [nic.to_dict() for nic in service.model.data.nics]
        mounts = {}
        for mount in service.model.data.mounts:
            fs_service = service.aysrepo.serviceGet('filesystem', mount.filesystem)
            try:
                sp = node.storagepools.get(fs_service.parent.name)
                fs = sp.get(fs_service.name)
            except KeyError:
                continue
            mounts[fs.path] = mount.target

        container = cls(
            name=service.name,
            node=node,
            mounts=mounts,
            nics=nics,
            hostname=service.model.data.hostname,
            flist=service.model.data.flist,
            ports=ports,
            host_network=service.model.data.hostNetworking,
            storage=service.model.data.storage,
            init_processes=[p.to_dict() for p in service.model.data.initProcesses],
            privileged=service.model.data.privileged,
            identity=service.model.data.identity,
        )
        return container

    @property
    def id(self):
        logger.debug("get container id")
        info = self.info
        if info:
            return info['container']['id']
        return

    @property
    def info(self):
        logger.debug("get container info")
        for containerid, container in self.node.client.container.list().items():
            if self.name in (container['container']['arguments']['tags'] or []):
                container['container']['id'] = int(containerid)
                return container
        return

    @property
    def client(self):
        if self._client is None:
            self._client = self.node.client.container.client(self.id)
        return self._client

    def upload_content(self, remote, content):
        if isinstance(content, str):
            content = content.encode('utf8')
        bytes = BytesIO(content)
        self.client.filesystem.upload(remote, bytes)

    def download_content(self, remote):
        buff = BytesIO()
        self.client.filesystem.download(remote, buff)
        return buff.getvalue().decode()
    
    def _create_container(self, timeout=60):
        logger.debug("send create container command to g8os")
        tags = [self.name]
        if self.hostname and self.hostname != self.name:
            tags.append(self.hostname)
        job = self.node.client.container.create(
            root_url=self.flist,
            mount=self.mounts,
            host_network=self.host_network,
            nics=self.nics,
            port=self.ports,
            tags=tags,
            hostname=self.hostname,
            storage=self.storage,
            privileged=self.privileged,
            identity=self.identity,
        )

        containerid = job.get(timeout)
        self._client = self.node.client.container.client(containerid)

    def is_job_running(self, cmd):
        try:
            for job in self._client.job.list():
                arguments = job['cmd']['arguments']
                if 'name' in arguments and arguments['name'] == cmd:
                    return job
            return False
        except Exception as err:
            if str(err).find("invalid container id"):
                return False
            raise

    def is_port_listening(self, port, timeout=60):
        import time
        start = time.time()
        while start + timeout > time.time():
            if port not in self.node.freeports(port, nrports=3):
                return True
            time.sleep(0.2)
        return False



    def start(self):
        if not self.is_running():
            logger.debug("start %s", self)
            self._create_container()
            for process in self.init_processes:
                cmd = "{} {}".format(process['name'], ' '.join(process.get('args', [])))
                pwd = process.get('pwd', '')
                stdin = process.get('stdin', '')
                env = {}
                for x in process.get('environment', []):
                    k, v = x.split("=")
                    env[k] = v
                self.client.system(command=cmd, dir=pwd, stdin=stdin, env=env)

    def stop(self):
        if not self.is_running():
            return
        logger.debug("stop %s", self)

        self.node.client.container.terminate(self.id)
        self._client = None

    def is_running(self):
        return self.id is not None

    @property
    def ays(self):
        if self._ays is None:
            from JumpScale.sal.g8os.atyourservice.StorageCluster import ContainerAYS
            self._ays = ContainerAYS(self)
        return self._ays

    def __str__(self):
        return "Container <{}>".format(self.name)

    def __repr__(self):
        return str(self)



from ..healthcheck import HealthCheckRun

descr = """
Check the bandwith consumption of the network
"""

class NetworkLoad(HealthCheckRun):
    def __init__(self, node):
        self.node = node
        resource = '/nodes/{}'.format(node.name)
        super().__init__(id='network-load', name="Network Load Check", category="Hardware", resource=resource)

    def run(self):
        results = []
        nics_speed = {nic['name']: nic['speed'] for nic in self.node.client.info.nic()}
        self.get_network_data('incoming', nics_speed)
        self.get_network_data('outgoing', nics_speed)

    def get_network_data(self, direction, nics_speed):
        throughput = self.node.client.aggregator.query('network.throughput.%s' % direction)
        for nic in throughput:
            throughput_history = throughput[nic].get('history', {}).get('3600', [])
            if throughput_history:
                last_throughput = throughput_history[-1].get('avg', 0)
                nic_name = nic.split("/")[-1]
                nic_speed = nics_speed.get(nic_name, 0)
                if nic_speed > 0:
                    nic_speed = nic_speed / 8
                    percent = (last_throughput/ float(nic_speed)) * 100
                    if percent > 90:
                        self.add_message(id="%s_%s" % (nic_name, direction), status="ERROR", text='Nic {} {} bandwith is {:.2f}%'.format(nic_name, direction, percent))
                    elif percent > 80:
                        self.add_message(id="%s_%s" % (nic_name, direction), status="WARNING", text='Nic {} {} bandwith is {:.2f}%'.format(nic_name, direction, percent))
                    else:
                        self.add_message(id="%s_%s" % (nic_name, direction), status="OK", text='Nic {} {} bandwith is {:.2f}%'.format(nic_name, direction, percent))



from js9 import j
import os
from zeroos.core0.client.client import Timeout
import json
import hashlib


class HealthCheckObject:
    def __init__(self, id, name, category, resource):
        self.id = id
        self.name = name
        self.category = category
        self._messages = []
        self.resource = resource
        self.stacktrace = ''

    def add_message(self, id, status, text):
        self._messages.append({'id': id, 'text': text, 'status': status})

    def to_dict(self):
        return {
            'id': self.id,
            'name': self.name,
            'category': self.category,
            'resource': self.resource,
            'stacktrace': self.stacktrace or '',
            'messages': self._messages
        }


class HealthCheckRun(HealthCheckObject):
    def start(self, *args, **kwargs):
        try:
            self.run(*args, **kwargs)
        except Exception as e:
            eco = j.errorhandler.parsePythonExceptionObject(e)
            self.stacktrace = eco.traceback
        return self.to_dict()


class IPMIHealthCheck(HealthCheckRun):
    def execute_ipmi(self, container, cmd):
        if self.node.client.filesystem.exists("/dev/ipmi") or self.node.client.filesystem.exists("/dev/ipmi0"):
            return container.client.system(cmd).get().stdout
        return ''


class ContainerContext:
    def __init__(self, node, flist):
        self.node = node
        self.flist = flist
        self.container = None
        self._name = 'healthcheck_{}'.format(hashlib.md5(flist.encode()).hexdigest())

    def __enter__(self):
        try:
            self.container = self.node.containers.get(self._name)
        except LookupError:
            self.container = self.node.containers.create(self._name, self.flist, host_network=True, privileged=True)
        return self.container

    def __exit__(self, exc_type, exc_val, exc_tb):
        return


class HealthCheck:
    def __init__(self, node):
        self.node = node
        self.healtcheckfolder = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'healthchecks')

    def with_container(self, flist):
        return ContainerContext(self.node, flist)

    def run(self, container, name, timeout=None):
        try:
            healthcheckfile = os.path.join(self.healtcheckfolder, name + '.py')
            if not os.path.exists(healthcheckfile):
                raise RuntimeError("Healtcheck with name {} not found".format(name))
            container.client.filesystem.upload_file('/tmp/{}.py'.format(name), healthcheckfile)
            try:
                job = container.client.bash('python3 /tmp/{}.py'.format(name))
                response = job.get(timeout)
            except Timeout:
                container.client.job.kill(job.id, 9)
                raise RuntimeError("Failed to execute {} on time".format(name))
            if response.state == 'ERROR':
                raise RuntimeError("Failed to execute {} {}".format(name, response.stdout))
            try:
                return json.loads(response.stdout)
            except Exception:
                raise RuntimeError("Failed to parse response of {}".format(name))
        except Exception as e:
            healtcheck = {
                'id': name,
                'status': 'ERROR',
                'message': str(e)
            }
            return healtcheck

    def cpu_mem(self):
        from .healthchecks.cpu_mem_core import CPU, Memory
        cpu = CPU(self.node)
        memory = Memory(self.node)
        return [cpu.start(), memory.start()]

    def disk_usage(self):
        from .healthchecks.diskusage import DiskUsage
        usage = DiskUsage(self.node)
        return usage.start()

    def network_bond(self):
        from .healthchecks.networkbond import NetworkBond
        bond = NetworkBond(self.node)
        return bond.start()

    def node_temperature(self, container):
        from .healthchecks.temperature import Temperature
        temperature = Temperature(self.node)
        result = temperature.start(container)
        return result

    def network_stability(self, nodes):
        from .healthchecks.networkstability import NetworkStability
        stability = NetworkStability(self.node)
        return stability.start(nodes)

    def rotate_logs(self):
        from .healthchecks.log_rotator import RotateLogs
        rotator = RotateLogs(self.node)
        return rotator.start()

    def openfiledescriptors(self):
        from .healthchecks.openfiledescriptors import OpenFileDescriptor
        ofd = OpenFileDescriptor(self.node)
        return ofd.start()

    def interrupts(self):
        from .healthchecks.interrupts import Interrupts
        inter = Interrupts(self.node)
        return inter.start()

    def threads(self):
        from .healthchecks.threads import Threads
        thread = Threads(self.node)
        return thread.start()

    def ssh_cleanup(self, job):
        from .healthchecks.ssh_cleanup import SSHCleanup
        cleaner = SSHCleanup(self.node, job)
        return cleaner.start()

    def powersupply(self, container):
        from .healthchecks.powersupply import PowerSupply
        powersupply = PowerSupply(self.node)
        return powersupply.start(container)

    def fan(self, container):
        from .healthchecks.fan import Fan
        fan = Fan(self.node)
        return fan.start(container)

    def context_switch(self):
        from .healthchecks.context_switch import ContextSwitch
        return ContextSwitch(self.node).start()

    def network_load(self):
        from .healthchecks.networkload import NetworkLoad
        load = NetworkLoad(self.node)
        return load.start()



"""
Auto-generated class for JobResult
"""
from .EnumJobResultName import EnumJobResultName
from .EnumJobResultState import EnumJobResultState

from . import client_support


class JobResult(object):
    """
    auto-generated. don't touch.
    """

    @staticmethod
    def create(data, id, level, name, startTime, state, stderr, stdout):
        """
        :type data: str
        :type id: str
        :type level: int
        :type name: EnumJobResultName
        :type startTime: int
        :type state: EnumJobResultState
        :type stderr: str
        :type stdout: str
        :rtype: JobResult
        """

        return JobResult(
            data=data,
            id=id,
            level=level,
            name=name,
            startTime=startTime,
            state=state,
            stderr=stderr,
            stdout=stdout,
        )

    def __init__(self, json=None, **kwargs):
        if json is None and not kwargs:
            raise ValueError('No data or kwargs present')

        class_name = 'JobResult'
        create_error = '{cls}: unable to create {prop} from value: {val}: {err}'
        required_error = '{cls}: missing required property {prop}'

        data = json or kwargs

        property_name = 'data'
        val = data.get(property_name)
        if val is not None:
            datatypes = [str]
            try:
                self.data = client_support.val_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))
        else:
            raise ValueError(required_error.format(cls=class_name, prop=property_name))

        property_name = 'id'
        val = data.get(property_name)
        if val is not None:
            datatypes = [str]
            try:
                self.id = client_support.val_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))
        else:
            raise ValueError(required_error.format(cls=class_name, prop=property_name))

        property_name = 'level'
        val = data.get(property_name)
        if val is not None:
            datatypes = [int]
            try:
                self.level = client_support.val_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))
        else:
            raise ValueError(required_error.format(cls=class_name, prop=property_name))

        property_name = 'name'
        val = data.get(property_name)
        if val is not None:
            datatypes = [EnumJobResultName]
            try:
                self.name = client_support.val_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))
        else:
            raise ValueError(required_error.format(cls=class_name, prop=property_name))

        property_name = 'startTime'
        val = data.get(property_name)
        if val is not None:
            datatypes = [int]
            try:
                self.startTime = client_support.val_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))
        else:
            raise ValueError(required_error.format(cls=class_name, prop=property_name))

        property_name = 'state'
        val = data.get(property_name)
        if val is not None:
            datatypes = [EnumJobResultState]
            try:
                self.state = client_support.val_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))
        else:
            raise ValueError(required_error.format(cls=class_name, prop=property_name))

        property_name = 'stderr'
        val = data.get(property_name)
        if val is not None:
            datatypes = [str]
            try:
                self.stderr = client_support.val_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))
        else:
            raise ValueError(required_error.format(cls=class_name, prop=property_name))

        property_name = 'stdout'
        val = data.get(property_name)
        if val is not None:
            datatypes = [str]
            try:
                self.stdout = client_support.val_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))
        else:
            raise ValueError(required_error.format(cls=class_name, prop=property_name))

    def __str__(self):
        return self.as_json(indent=4)

    def as_json(self, indent=0):
        return client_support.to_json(self, indent=indent)

    def as_dict(self):
        return client_support.to_dict(self)



from zerotier.client import Client
import netaddr


class ZTBootstrap:

    def __init__(self, token, bootstap_id, grid_id, cidr):
        self.bootstap_nwid = bootstap_id
        self.grid_nwid = grid_id
        self._cidr = cidr  # TODO validate format
        # create client and set the authentication header
        self._zt = Client()
        self._zt.set_auth_header("Bearer " + token)

    def configure_routes(self):
        for nwid in [self.bootstap_nwid, self.grid_nwid]:
            resp = self._zt.network.getNetwork(nwid)
            resp.raise_for_status()
            nw = resp.json()
            nw['config']['routes'] = [{'target': self._cidr, 'via': None}]
            self._zt.network.updateNetwork(nw, nwid).raise_for_status()

    def list_join_request(self):
        """
        return a list of member that try to access the bootstap network
        """
        resp = self._zt.network.listMembers(id=self.bootstap_nwid)
        resp.raise_for_status()

        requests = []
        for member in resp.json():
            if not member['online'] or member['config']['authorized']:
                continue
            requests.append(member)

        return requests

    def assign_ip(self, nwid, member, ip=None):
        """
        Assign an Ip address to a member in a certain network
        @nwid : id of the network
        @member : member object
        @ip: ip address to assing to the member, if None take the next free IP in the range
        """
        if ip is None:
            ip = self._find_free_ip(nwid)
        member['config']['authorized'] = True
        member['config']['ipAssignments'] = [ip]
        resp = self._zt.network.updateMember(member, member['nodeId'], nwid)
        resp.raise_for_status()
        return ip

    def unauthorize_member(self, nwid, member):
        member['config']['authorized'] = False
        member['config']['ipAssignments'] = []
        resp = self._zt.network.updateMember(member, member['nodeId'], nwid)
        resp.raise_for_status()


    def _find_free_ip(self, nwid):

        resp = self._zt.network.listMembers(nwid)
        resp.raise_for_status()

        all_ips = list(netaddr.IPNetwork(self._cidr))
        for member in resp.json():
            for addr in member['config']['ipAssignments']:
                all_ips.remove(netaddr.IPAddress(addr))
        if len(all_ips) <= 0:
            raise RuntimeError("No more free ip in the range %s" % self._cidr)
        return str(all_ips[0])


if __name__ == '__main__':
    token = '4gE9Cfqw2vFFzCPC1BYaj2mbSpNScxJx'
    bootstap_nwid = '17d709436c993670'
    grid_nwid = 'a09acf02336ce8b5'

    zt = ZTBootstrap(token, bootstap_nwid, grid_nwid, '192.168.10.0/24')
    from IPython import embed; embed()



"""
Auto-generated class for OSInfo
"""

from . import client_support


class OSInfo(object):
    """
    auto-generated. don't touch.
    """

    @staticmethod
    def create(bootTime, hostname, os, platform, platformFamily, platformVersion, procs, uptime, virtualizationRole, virtualizationSystem):
        """
        :type bootTime: int
        :type hostname: str
        :type os: str
        :type platform: str
        :type platformFamily: str
        :type platformVersion: str
        :type procs: int
        :type uptime: int
        :type virtualizationRole: str
        :type virtualizationSystem: str
        :rtype: OSInfo
        """

        return OSInfo(
            bootTime=bootTime,
            hostname=hostname,
            os=os,
            platform=platform,
            platformFamily=platformFamily,
            platformVersion=platformVersion,
            procs=procs,
            uptime=uptime,
            virtualizationRole=virtualizationRole,
            virtualizationSystem=virtualizationSystem,
        )

    def __init__(self, json=None, **kwargs):
        if json is None and not kwargs:
            raise ValueError('No data or kwargs present')

        class_name = 'OSInfo'
        create_error = '{cls}: unable to create {prop} from value: {val}: {err}'
        required_error = '{cls}: missing required property {prop}'

        data = json or kwargs

        property_name = 'bootTime'
        val = data.get(property_name)
        if val is not None:
            datatypes = [int]
            try:
                self.bootTime = client_support.val_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))
        else:
            raise ValueError(required_error.format(cls=class_name, prop=property_name))

        property_name = 'hostname'
        val = data.get(property_name)
        if val is not None:
            datatypes = [str]
            try:
                self.hostname = client_support.val_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))
        else:
            raise ValueError(required_error.format(cls=class_name, prop=property_name))

        property_name = 'os'
        val = data.get(property_name)
        if val is not None:
            datatypes = [str]
            try:
                self.os = client_support.val_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))
        else:
            raise ValueError(required_error.format(cls=class_name, prop=property_name))

        property_name = 'platform'
        val = data.get(property_name)
        if val is not None:
            datatypes = [str]
            try:
                self.platform = client_support.val_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))
        else:
            raise ValueError(required_error.format(cls=class_name, prop=property_name))

        property_name = 'platformFamily'
        val = data.get(property_name)
        if val is not None:
            datatypes = [str]
            try:
                self.platformFamily = client_support.val_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))
        else:
            raise ValueError(required_error.format(cls=class_name, prop=property_name))

        property_name = 'platformVersion'
        val = data.get(property_name)
        if val is not None:
            datatypes = [str]
            try:
                self.platformVersion = client_support.val_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))
        else:
            raise ValueError(required_error.format(cls=class_name, prop=property_name))

        property_name = 'procs'
        val = data.get(property_name)
        if val is not None:
            datatypes = [int]
            try:
                self.procs = client_support.val_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))
        else:
            raise ValueError(required_error.format(cls=class_name, prop=property_name))

        property_name = 'uptime'
        val = data.get(property_name)
        if val is not None:
            datatypes = [int]
            try:
                self.uptime = client_support.val_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))
        else:
            raise ValueError(required_error.format(cls=class_name, prop=property_name))

        property_name = 'virtualizationRole'
        val = data.get(property_name)
        if val is not None:
            datatypes = [str]
            try:
                self.virtualizationRole = client_support.val_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))
        else:
            raise ValueError(required_error.format(cls=class_name, prop=property_name))

        property_name = 'virtualizationSystem'
        val = data.get(property_name)
        if val is not None:
            datatypes = [str]
            try:
                self.virtualizationSystem = client_support.val_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))
        else:
            raise ValueError(required_error.format(cls=class_name, prop=property_name))

    def __str__(self):
        return self.as_json(indent=4)

    def as_json(self, indent=0):
        return client_support.to_json(self, indent=indent)

    def as_dict(self):
        return client_support.to_dict(self)



class HealthService:
    def __init__(self, client):
        self.client = client



    def ListNodeHealth(self, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Get detailed information of a node  health
        It is method for GET /health/nodes/{nodeid}
        """
        uri = self.client.base_url + "/health/nodes/"+nodeid
        return self.client.get(uri, None, headers, query_params, content_type)


    def ListNodesHealth(self, headers=None, query_params=None, content_type="application/json"):
        """
        List all nodes health
        It is method for GET /health/nodes
        """
        uri = self.client.base_url + "/health/nodes"
        return self.client.get(uri, None, headers, query_params, content_type)



from ..healthcheck import HealthCheckRun

descr = """
Monitors if threads has high number of threads per hyperthread
"""


class Threads(HealthCheckRun):
    def __init__(self, node):
        resource = '/nodes/{}'.format(node.name)
        super().__init__(id="threads",
                         name="Node thread per hyperthread",
                         category="System Load",
                         resource=resource)
        self.node = node

    def run(self):
        hyperthread_count = len(self.node.client.info.cpu())

        thread_info = self.node.client.aggregator.query("machine.process.threads", type="phys")
        thread_info = thread_info.get("machine.process.threads", {})
        message = {}
        if not thread_info:
            message = {
                "status": "WARNING",
                "id": "THREADS",
                "text": "Number of threads is not available",
            }
            self.add_message(**message)
            return

        avg_thread = thread_info["current"]["300"]["avg"] / hyperthread_count
        message["id"] = "THREADS"
        if avg_thread > 300:
            message["status"] = "WARNING"
            message["text"] = "Average threads per hyperthread is high"
        elif avg_thread > 400:
            message["status"] = "ERROR"
            message["text"] = "Average threads per hyperthread is critical"
        else:
            message["status"] = "OK"
            message["text"] = "Average threads per hyperthread is normal"

        self.add_message(**message)



import time
from ..healthcheck import HealthCheckRun

descr = """
Monitors if a number of interrupts
"""


class Interrupts(HealthCheckRun):
    def __init__(self, node, warn=8000, error=1000):
        resource = '/nodes/{}'.format(node.name)
        super().__init__('interrupts', 'CPU Interrupts', 'Hardware', resource)
        self._warn = warn
        self._error = error
        self.node = node

    def _get(self):
        client = self.node.client

        state = client.aggregator.query('machine.CPU.interrupts').get('machine.CPU.interrupts')
        if state is None:
            # nothing to check yet
            return {
                'id': self.id,
                'status': 'WARNING',
                'text': 'Number of interrupts per second is not collected yet',
            }

        # time of last reported value
        last_time = state['last_time']
        current = state['current']['300']
        # start time of the current 5min sample
        current_time = current['start']
        if current_time < time.time() - (10*60):
            return {
                'id': self.id,
                'status': 'WARNING',
                'text': 'Last collected interrupts are too far in the past',
            }

        # calculate avg per second
        value = current['avg'] / (last_time - current_time)

        status = 'OK'
        text = 'Interrupts are okay'

        if value >= self._error:
            status = 'ERROR'
            text = 'Interrupts exceeded error threshold of {} ({})'.format(self._error, value)
        elif value >= self._warn:
            status = 'WARNING'
            text = 'Interrupts exceeded warning threshold of {} ({})'.format(self._warn, value)

        return {
            'id': self.id,
            'status': status,
            'text': text,
        }

    def run(self):
        self.add_message(**self._get())



from enum import Enum
from .Partition import Partition
from .abstracts import Mountable


class DiskType(Enum):
    ssd = "ssd"
    hdd = "hdd"
    nvme = "nvme"
    archive = "archive"
    cdrom = 'cdrom'


class Disks:

    """Subobject to list disks"""
    def __init__(self, node):
        self.node = node
        self._client = node.client

    def list(self):
        """
        List of disks on the node
        """
        disks = []
        disk_list = self._client.disk.list()
        if 'blockdevices' in disk_list:
            for disk_info in self._client.disk.list()['blockdevices']:
                disks.append(Disk(
                    node=self.node,
                    disk_info=disk_info
                ))
        return disks

    def get(self, name):
        """
        return the disk called `name`
        @param name: name of the disk
        """
        for disk in self.list():
            if disk.name == name:
                return disk
        return None


class Disk(Mountable):
    """Disk in a G8OS"""

    def __init__(self, node, disk_info):
        """
        disk_info: dict returned by client.disk.list()
        """
        # g8os client to talk to the node
        self.node = node
        self._client = node.client
        self.name = None
        self.size = None
        self.blocksize = None
        self.partition_table = None
        self.mountpoint = None
        self.model = None
        self._filesystems = []
        self.type = None
        self.partitions = []

        self._load(disk_info)

    @property
    def devicename(self):
        return "/dev/{}".format(self.name)

    @property
    def filesystems(self):
        self._populate_filesystems()
        return self._filesystems

    def _load(self, disk_info):
        self.name = disk_info['name']
        detail = self._client.disk.getinfo(self.name)
        self.size = int(disk_info['size'])
        self.blocksize = detail['blocksize']
        if detail['table'] != 'unknown':
            self.partition_table = detail['table']
        self.mountpoint = disk_info['mountpoint']
        self.model = disk_info['model']
        self.type = self._disk_type(disk_info)
        for partition_info in disk_info.get('children', []) or []:
            self.partitions.append(
                Partition(
                    disk=self,
                    part_info=partition_info)
            )

    def _populate_filesystems(self):
        """
        look into all the btrfs filesystem and populate
        the filesystems attribute of the class with the detail of
        all the filesystem present on the disk
        """
        self._filesystems = []
        for fs in (self._client.btrfs.list() or []):
            for device in fs['devices']:
                if device['path'] == "/dev/{}".format(self.name):
                    self._filesystems.append(fs)
                    break

    def _disk_type(self, disk_info):
        """
        return the type of the disk
        """
        if disk_info['rota'] == "1":
            if disk_info['type'] == 'rom':
                return DiskType.cdrom
            # assume that if a disk is more than 7TB it's a SMR disk
            elif int(disk_info['size']) > (1024 * 1024 * 1024 * 1024 * 7):
                return DiskType.archive
            else:
                return DiskType.hdd
        else:
            if "nvme" in disk_info['name']:
                return DiskType.nvme
            else:
                return DiskType.ssd

    def mktable(self, table_type='gpt', overwrite=False):
        """
        create a partition table on the disk
        @param table_type: Partition table type as accepted by parted
        @param overwrite: erase any existing partition table
        """
        if self.partition_table is not None and overwrite is False:
            return

        self._client.disk.mktable(
            disk=self.name,
            table_type=table_type
        )

    def mkpart(self, start, end, part_type="primary"):
        """
        @param start: partition start as accepted by parted mkpart
        @param end: partition end as accepted by parted mkpart
        @param part_type: partition type as accepted by parted mkpart
        """
        before = {p.name for p in self.partitions}

        self._client.disk.mkpart(
            self.name,
            start=start,
            end=end,
            part_type=part_type,
        )
        after = {}
        for disk in self._client.disk.list()['blockdevices']:
            if disk['name'] != self.name:
                continue
            for part in disk.get('children', []):
                after[part['name']] = part
        name = set(after.keys()) - before

        part_info = after[list(name)[0]]
        partition = Partition(
            disk=self,
            part_info=part_info)
        self.partitions.append(partition)

        return partition

    def __str__(self):
        return "Disk <{}>".format(self.name)

    def __repr__(self):
        return str(self)

    def __eq__(self, other):
        return self.devicename == other.devicename



"""
Auto-generated class for DiskInfo
"""
from .DiskPartition import DiskPartition
from .EnumDiskInfoType import EnumDiskInfoType

from . import client_support


class DiskInfo(object):
    """
    auto-generated. don't touch.
    """

    @staticmethod
    def create(device, partitions, size, type):
        """
        :type device: str
        :type partitions: list[DiskPartition]
        :type size: int
        :type type: EnumDiskInfoType
        :rtype: DiskInfo
        """

        return DiskInfo(
            device=device,
            partitions=partitions,
            size=size,
            type=type,
        )

    def __init__(self, json=None, **kwargs):
        if json is None and not kwargs:
            raise ValueError('No data or kwargs present')

        class_name = 'DiskInfo'
        create_error = '{cls}: unable to create {prop} from value: {val}: {err}'
        required_error = '{cls}: missing required property {prop}'

        data = json or kwargs

        property_name = 'device'
        val = data.get(property_name)
        if val is not None:
            datatypes = [str]
            try:
                self.device = client_support.val_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))
        else:
            raise ValueError(required_error.format(cls=class_name, prop=property_name))

        property_name = 'partitions'
        val = data.get(property_name)
        if val is not None:
            datatypes = [DiskPartition]
            try:
                self.partitions = client_support.list_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))
        else:
            raise ValueError(required_error.format(cls=class_name, prop=property_name))

        property_name = 'size'
        val = data.get(property_name)
        if val is not None:
            datatypes = [int]
            try:
                self.size = client_support.val_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))
        else:
            raise ValueError(required_error.format(cls=class_name, prop=property_name))

        property_name = 'type'
        val = data.get(property_name)
        if val is not None:
            datatypes = [EnumDiskInfoType]
            try:
                self.type = client_support.val_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))
        else:
            raise ValueError(required_error.format(cls=class_name, prop=property_name))

    def __str__(self):
        return self.as_json(indent=4)

    def as_json(self, indent=0):
        return client_support.to_json(self, indent=indent)

    def as_dict(self):
        return client_support.to_dict(self)



import time
from ..healthcheck import HealthCheckRun

descr = """
Monitors CPU context switching
"""


class ContextSwitch(HealthCheckRun):
    def __init__(self, node, warn=600000, error=1000000):
        super().__init__('context-switch', 'Context Switch', 'Hardware', '/nodes/{}'.format(node.name))
        self.node = node
        self._warn = warn
        self._error = error

    def run(self):
        client = self.node.client

        state = client.aggregator.query('machine.CPU.contextswitch').get('machine.CPU.contextswitch')
        if state is None:
            # nothing to check yet
            return self.add_message(self.id, 'WARNING', 'Number of context-switches per second is not collected yet')

        # time of last reported value
        last_time = state['last_time']
        current = state['current']['300']
        # start time of the current 5min sample
        current_time = current['start']
        if current_time < time.time() - (10*60):
            return self.add_message(self.id, 'WARNING', 'Last collected context-switch are too far in the past')

        # calculate avg per second
        value = current['avg'] / (last_time - current_time)

        status = 'OK'
        text = 'Context-switches are okay'

        if value >= self._error:
            status = 'ERROR'
            text = 'Contex-switches exceeded error threshold of {} ({})'.format(self._error, value)
        elif value >= self._warn:
            status = 'WARNING'
            text = 'Contex-switches exceeded warning threshold of {} ({})'.format(self._warn, value)

        return self.add_message(self.id, status, text)



from .abstracts import Mountable


class Partition(Mountable):
    """Partition of a disk in a G8OS"""

    def __init__(self, disk, part_info):
        """
        part_info: dict returned by client.disk.list()
        """
        # g8os client to talk to the node
        self.disk = disk
        self._client = disk.node.client
        self.name = None
        self.size = None
        self.blocksize = None
        self.mountpoint = None
        self.uuid = None
        self._filesystems = []

        self._load(part_info)

    @property
    def filesystems(self):
        self._populate_filesystems()
        return self._filesystems

    @property
    def devicename(self):
        return "/dev/{}".format(self.name)

    def _load(self, part_info):
        self.name = part_info['name']
        self.size = int(part_info['size'])
        self.blocksize = self.disk.blocksize
        self.mountpoint = part_info['mountpoint']
        self.uuid = part_info['partuuid']

    def _populate_filesystems(self):
        """
        look into all the btrfs filesystem and populate
        the filesystems attribute of the class with the detail of
        all the filesystem present on the disk
        """
        self._filesystems = []
        for fs in (self._client.btrfs.list() or []):
            for device in fs['devices']:
                if device['path'] == "/dev/{}".format(self.name):
                    self._filesystems.append(fs)
                    break

    def __str__(self):
        return "Partition <{}>".format(self.name)

    def __repr__(self):
        return str(self)



from ..healthcheck import IPMIHealthCheck

descr = """
Checks temperature of the system.
Result will be shown in the "Temperature" section of the Grid Portal / Status Overview / Node Status page.
"""


class Temperature(IPMIHealthCheck):
    WARNING_TRIPPOINT = 70
    ERROR_TRIPPOINT = 90

    def __init__(self, node):
        self.node = node
        resource = '/nodes/{}'.format(node.name)
        super().__init__('temperature', 'Node Temperature Check', 'Hardware', resource)

    def run(self, container):
        out = self.execute_ipmi(container, "ipmitool sdr type 'Temp'")
        if out:
            if out:
                # SAMPLE:
                # root@du-conv-3-01:~# ipmitool sdr type "Temp"
                # Temp             | 0Eh | ok  |  3.1 | 37 degrees C
                # Temp             | 0Fh | ok  |  3.2 | 34 degrees C
                # Inlet Temp       | B1h | ok  | 64.96 | 28 degrees C
                # Exhaust Temp     | B2h | ns  | 144.96 | Disabled
                for line in out.splitlines():
                    if "|" in line:
                        parts = [part.strip() for part in line.split("|")]
                        id_, sensorstatus, message = parts[0], parts[2], parts[-1]

                        if sensorstatus == "ns" and "no reading" in message.lower():
                            continue

                        if sensorstatus != "ok" and "no reading" not in message.lower():
                            self.add_message(**self.get_message(sensor=id_, status='WARNING', message=message))
                            continue
                        temperature = int(message.split(" ", 1)[0])
                        self.add_message(**self.get_message(sensor=id_, status=sensorstatus, message=message, temperature=temperature))
        else:
            self.add_message(**self.get_message(status="SKIPPED", message="NO temp information available"))

    def get_message(self, sensor="", status='OK', message='', temperature=0):
        result = {
            "status": status.upper(),
            "text": "%s: %s" % (sensor, message),
            "id": sensor,
        }
        if status != "OK":
            return result

        if temperature >= self.WARNING_TRIPPOINT:
            result["status"] = "WARNING"
        elif temperature >= self.ERROR_TRIPPOINT:
            result["status"] = "ERROR"
        return result



from ..healthcheck import IPMIHealthCheck
descr = """
Checks the fans of a node using IPMItool.
Result will be shown in the "Hardware" section of the Grid Portal / Status Overview / Node Status page.
"""


class Fan(IPMIHealthCheck):
    def __init__(self, node):
        self.node = node
        resource = '/nodes/{}'.format(node.name)
        super().__init__(id='fan', name="Fan", category="Hardware", resource=resource)

    def run(self, container):
        out = self.execute_ipmi(container, """ipmitool sdr type "Fan" """)
        if out:
            # SAMPLE:
            # root@du-conv-3-01:~# ipmitool sdr type "Fan"
            # FAN1             | 41h | ok  | 29.1 | 5000 RPM
            # FAN2             | 42h | ns  | 29.2 | No Reading
            # FAN3             | 43h | ok  | 29.3 | 4800 RPM
            # FAN4             | 44h | ns  | 29.4 | No Reading

            for line in out.splitlines():
                parts = [part.strip() for part in line.split("|")]
                id_, sensorstatus, text = parts[0], parts[2], parts[-1]
                if sensorstatus == "ns" and "no reading" in text.lower():
                    self.add_message(id=id_, status='SKIPPED', text="Fan {id} has no reading ({text})".format(id=id_, text=text))
                elif sensorstatus != "ok" and "no reading" not in text.lower():
                    self.add_message(id=id_, status='WARNING', text="Fan {id} has problem ({text})".format(id=id_, text=text))
                elif sensorstatus == 'ok':
                    self.add_message(id=id_, status="OK", text="Fan {id} is working at ({text})".format(id=id_, text=text))
        else:
            self.add_message(id="SKIPPED", status="SKIPPED", text="NO fan information available")
